import os
import json

from legacy.short_pka_sets import pKa_sets_short as pka_sets_short


def convert_short_set_to_original_json(pka_sets, name,
                                       output_filepath):
    """
    Converts the legacy data into JSON
    without applying any further processing.

    """
    short_set = pka_sets[name]
    new_short_set = {
        "name": name,
        "values": short_set
    }
    with open(output_filepath, "w") as f:
        json.dump(new_short_set, f, indent=2)


if __name__ == '__main__':
    # Get parent dir where the script runs
    parent_dir = os.path.dirname(os.path.realpath(__file__))

    # Converts each data set from pka_sets_short into JSON
    for name in pka_sets_short.keys():
        filepath = f"{parent_dir}/original_short/" \
            f"pka_set_{name.lower()}_original.json"
        convert_short_set_to_original_json(pka_sets_short, name,
                                           filepath)



### Calibrated ExPASY - from Gauci et al. Proteomics 2008, 8, 4898 as implemented in pIR
SetName='Gauci'

# Acidic_Amino_Acids
#             AA    Primary  N-Terminal  C-Terminal
acidic1 = { 
 "D": [ 4.05, 4.05, 4.05 ],
 "C": [ 9.0 , 9.0 , 9.0  ],
 "E": [ 4.45, 4.45, 4.45 ],
 "Y": [ 10.0, 10.0, 10.0 ], 
 'U': [ 5.43, 5.20, 5.60 ]  # pK for U was taken from Byun et al. Biopolymers 2011, 95, 345
}


# Basic_Amino_Acids
#             AA    Primary  N-Terminal  C-Terminal
basic1 = { 
 "R": [ 12.0, 12.0, 12.0 ],
 "H": [ 5.98, 5.98, 5.98 ],
 "K": [ 10.0, 10.0, 10.0 ]
}


# Terminal_Amino_Acids
# AA N-term  C-Term
terminus_ionizable1 = { 
 "A": [  7.59,   3.55 ], 
 "R": [  7.5,    3.55 ], 
 "N": [  6.7,    3.55 ], 
 "D": [  7.5,    4.55 ], 
 "C": [  6.5,    3.55 ], 
 "E": [  7.7,    4.75 ], 
 "Q": [  7.5,    3.55 ], 
 "G": [  7.5,    3.55 ], 
 "H": [  7.5,    3.55 ], 
 "I": [  7.5,    3.55 ], 
 "L": [  7.5,    3.55 ], 
 "K": [  7.5,    3.55 ], 
 "M": [  7.0,    3.55 ], 
 "F": [  7.5,    3.55 ], 
 "P": [  8.3599, 3.55 ],
 "S": [  6.93,   3.55 ], 
 "T": [  6.82,   3.55 ], 
 "W": [  7.5,    3.55 ], 
 "Y": [  7.5,    3.55 ], 
 "V": [  7.44,   3.55 ]   
}


known_res=['G', 'A', 'S', 'P', 'V', 'T', 'C', 'I', 'L', 'N', 'D', 'Q', 'K', 'E', 'M', 'H', 'F', 'R', 'Y', 'W', 'X', 'Z', 'B', 'U']


def FillMissingAAtoterminus_ionizable(terminus_ionizable):
    
    # Calc average
    sumNterm=0
    sumCterm=0
    #for k,v in terminus_ionizable.iteritems():
    for k in terminus_ionizable.keys():
        v = terminus_ionizable[k]
        sumNterm += v[0]
        sumCterm += v[1]
    avNterm = sumNterm / len(terminus_ionizable.keys())
    avCterm = sumCterm / len(terminus_ionizable.keys())

    for R in known_res:
        if R not in terminus_ionizable.keys():
            if   R == 'X': terminus_ionizable[R] = [ avNterm, avCterm ]
            elif R == 'Z': terminus_ionizable[R] = [ (terminus_ionizable['E'][0]+terminus_ionizable['Q'][0])/2,  (terminus_ionizable['E'][1]+terminus_ionizable['Q'][1])/2  ]
            elif R == 'B': terminus_ionizable[R] = [ (terminus_ionizable['N'][0]+terminus_ionizable['D'][0])/2,  (terminus_ionizable['N'][1]+terminus_ionizable['D'][1])/2  ]
            elif R == 'U': 
                # copy of X
                terminus_ionizable[R] = [ avNterm, avCterm ]
            else:
                print("---!Error: data for specific -NH2 and -COOH termini pKa values for residue "+R+" is not given in the "+SetName+" pKa set. Set this residue identical to X (average of all available). Check set. Exit.")
                # sys.exit(1)
    
    return terminus_ionizable


from pichemist.charges import PKaChargeCalculator
from pichemist.fasta.matcher import FastaPKaMatcher
from pichemist.isoelectric import CurveCalculator
from pichemist.isoelectric import IsoelectricCalculator
from pichemist.model import OutputFragAttribute
from pichemist.molecule import smiles_to_image
from pichemist.pka.utils import merge_pkas_lists
from pichemist.stats import mean
from pichemist.stats import stddev
from pichemist.stats import stderr


def get_net_qs_from_qs_and_frags(net_qs_and_frags):
    """Returns a list of charges from charge-fragment tuples."""
    return [v[0] for v in net_qs_and_frags]


def generate_stats_on_dict(input_dict, mean_title="mean"):
    """
    Enriches the an input dictionary of float values
    with mean, std deviation, and std error.

    """
    value_list = list()
    for k in input_dict.keys():
        value_list += [input_dict[k]]
    input_dict[mean_title] = mean(value_list)
    input_dict["std"] = stddev(value_list)
    input_dict["err"] = stderr(value_list)
    return input_dict


def get_low_and_high_from_interval_lists(interval_low_list, interval_high_list):
    """
    Processes interval lists to yield their averages.
    If molecule permanently has a charge, the interval is
    not defined and NaN are provided instead.

    """
    if len(interval_low_list) > 0:
        interval_low = mean(interval_low_list)
    else:
        interval_low = float("NaN")
    if len(interval_high_list) > 0:
        interval_high = mean(interval_high_list)
    else:
        interval_high = float("NaN")
    return interval_low, interval_high


def merge_matched_and_calculated_pkas(
    base_pkas_fasta, base_pkas_calc, acid_pkas_fasta, acid_pkas_calc
):
    """Merge FASTA-matched and calculated pKas."""
    base_pkas_list = [base_pkas_fasta, base_pkas_calc]
    acid_pkas_list = [acid_pkas_fasta, acid_pkas_calc]
    base_pkas_dict, acid_pkas_dict = merge_pkas_lists(base_pkas_list, acid_pkas_list)
    return base_pkas_dict, acid_pkas_dict


def calculate_pI_pH_and_charge_dicts(base_pkas_dict, acid_pkas_dict, net_qs_and_frags):
    """Calculates the isoelectric point, charges, and pH charges."""
    pI_dict = dict()
    q_dict = dict()
    pH_q_dict = dict()
    pka_sets_names = FastaPKaMatcher().get_pka_sets_names()
    for pka_set in pka_sets_names:

        # Merge fasta and calculated pkas
        base_pkas = base_pkas_dict[pka_set]
        acid_pkas = acid_pkas_dict[pka_set]

        # Calculate isoelectric point
        net_qs = get_net_qs_from_qs_and_frags(net_qs_and_frags)
        constant_q = PKaChargeCalculator().calculate_constant_charge(net_qs)
        q = PKaChargeCalculator().calculate_charge(
            base_pkas, acid_pkas, pH=7.4, constant_q=constant_q
        )
        pI = IsoelectricCalculator().calculate_pI(
            base_pkas, acid_pkas, constant_q=constant_q
        )
        pH_q = CurveCalculator().calculate_charged_curve(
            base_pkas, acid_pkas, constant_q=constant_q
        )
        pI_dict[pka_set] = pI
        q_dict[pka_set] = q
        pH_q_dict[pka_set] = pH_q

    # Generate stats
    pI_dict = generate_stats_on_dict(pI_dict, mean_title="pI mean")
    q_dict = generate_stats_on_dict(q_dict, mean_title="Q at pH7.4 mean")
    return pI_dict, q_dict, pH_q_dict


def calculate_isoelectric_interval_and_threshold(pH_q_dict):
    """
    Calculates the isoelectric interval that is
    the pH range where the charge is within the given threshold.

    """
    threshold = IsoelectricCalculator().interval_threshold
    interval_low_list = list()
    interval_high_list = list()

    # Calculate intervals across all sets
    for pka_set in FastaPKaMatcher().get_pka_sets_names():
        pH_int = IsoelectricCalculator().calculate_interval(pH_q_dict[pka_set])
        if len(pH_int) > 1:
            interval_low_list.append(pH_int[0])
            interval_high_list.append(pH_int[-1])

    # Average and return results
    interval_low, interval_high = get_low_and_high_from_interval_lists(
        interval_low_list, interval_high_list
    )
    return (interval_low, interval_high), threshold


def compile_frags_pkas_for_output(
    base_pkas_fasta,
    acid_pkas_fasta,
    base_pkas_calc,
    acid_pkas_calc,
    net_qs_and_frags,
    generate_fragment_images=False,
):
    """
    Produces dictionary with fragments (known AA or SMILES fragment),
    their occurences in the molecule, and corresponding pKa
    (average between pKa sets in the case of known AA).

    """
    frag_acid_pkas_fasta = calculate_frags_for_output_fasta("acid", acid_pkas_fasta)
    frag_base_pkas_fasta = calculate_frags_for_output_fasta("base", base_pkas_fasta)
    frag_acid_pkas_calc = calculate_frags_for_output_calc(
        "acid",
        acid_pkas_calc,
        generate_fragment_images=generate_fragment_images,
    )
    frag_base_pkas_calc = calculate_frags_for_output_calc(
        "base",
        base_pkas_calc,
        generate_fragment_images=generate_fragment_images,
    )
    frag_Qs_calc = calculate_frags_for_output_calc(
        "constant charge",
        net_qs_and_frags,
        generate_fragment_images=generate_fragment_images,
    )
    return (
        frag_acid_pkas_fasta,
        frag_base_pkas_fasta,
        frag_acid_pkas_calc,
        frag_base_pkas_calc,
        frag_Qs_calc,
    )


def calculate_frags_for_output_calc(
    ionization_type, pkas_calc, generate_fragment_images=False
):
    """
    Generates a dictionary of fragments with their pKas.
    Images can additionally be included.

    """
    frag_pkas_calc = dict()
    frg_idx = 0
    for v in pkas_calc:
        frg_idx += 1
        pka = v[0]
        smi = v[1]

        frag_pkas_calc[frg_idx] = {
            OutputFragAttribute.TYPE: ionization_type,
            OutputFragAttribute.FRAGMENT: smi,
            OutputFragAttribute.COUNT: 1,
            OutputFragAttribute.PKA: pka,
        }

        if generate_fragment_images:
            frag_pkas_calc[frg_idx][OutputFragAttribute.IMAGE] = smiles_to_image(smi)
    return frag_pkas_calc


def calculate_frags_for_output_fasta(ionization_type, pkas_fasta):
    """Generates a dictionary of fragments with their pKas."""
    pka_dict = dict()
    count_dict = dict()
    pka_sets_count = 0
    for pka_set, list_for_pka_set in pkas_fasta.items():
        pka_sets_count += 1
        for v in list_for_pka_set:
            pka = v[0]
            aa = v[1]
            if aa in pka_dict.keys():
                pka_dict[aa].append(pka)
            else:
                pka_dict[aa] = list()
            if pka_sets_count == 1:
                if aa in count_dict.keys():
                    count_dict[aa] += 1
                else:
                    count_dict[aa] = 1

    frag_pkas_fasta = dict()
    idx = 0
    for k, v in pka_dict.items():
        idx += 1
        frag_pkas_fasta[idx] = {
            OutputFragAttribute.TYPE: ionization_type,
            OutputFragAttribute.FRAGMENT: k,
            OutputFragAttribute.COUNT: count_dict[k],
            OutputFragAttribute.PKA: sum(v) / len(v),
        }
    return frag_pkas_fasta



import os

from legacy import gauci
from legacy_conversion import convert_legacy_set_to_json


def build_gauci_set_from_module(gm: gauci,
                                fill_missing_aas=False):
    """Builds a pKa set from a module."""
    pka_set = {
        "acidic": gm.acidic1,
        "basic": gm.basic1,
        "terminus_ionizable": gm.terminus_ionizable1
    }
    if fill_missing_aas:
        pka_set = _fill_missing_terminus_ionizable_aas(pka_set)
    return pka_set


def _fill_missing_terminus_ionizable_aas(pka_set):
    """Adds the missing terminus ionizable AAs."""
    pka_set["terminus_ionizable"] = \
        gauci.FillMissingAAtoterminus_ionizable(
            pka_set["terminus_ionizable"]
        )
    return pka_set


if __name__ == '__main__':
    # Get parent dir where the script runs
    parent_dir = os.path.dirname(os.path.realpath(__file__))

    # Converts the original legacy data set into JSON
    name = gauci.SetName
    pka_set = build_gauci_set_from_module(gauci, fill_missing_aas=False)
    filepath = f"{parent_dir}/original_enriched/" \
        f"pka_set_{name.lower()}_original.json"
    pka_dict = convert_legacy_set_to_json(pka_set, name, filepath)

    # Converts the legacy data set with filled missing AAs into JSON
    pka_set = build_gauci_set_from_module(gauci, fill_missing_aas=True)
    filepath = f"{parent_dir}/standardised/" \
        f"pka_set_{name.lower()}_standardised.json"
    pka_dict = convert_legacy_set_to_json(pka_set, name, filepath)



import numpy as np
from pichemist.charges import PKaChargeCalculator
from pichemist.utils import get_logger

log = get_logger(__name__)


class CurveCalculator(object):
    """Uses pKa values to calculate the pH/Q curve."""

    def __init__(self):
        self.pH_lower_bound = -1
        self.pH_upper_bound = 15
        self.pH_step = 0.1
        self.pH_range = self._define_pH_range()
        self.q_range = self._define_charge_range()

    def _define_pH_range(self):
        """Sets the pH range for the curve (X axis)."""
        pH_limit = self.get_pH_span()
        return np.arange(pH_limit[0], pH_limit[1] + self.pH_step, self.pH_step)

    def _define_charge_range(self):
        """Sets the charge range (Y axis)."""
        return self.pH_range * 0.0

    def get_pH_span(self):
        """
        Define pH span to, for example, calculate the
        pH/Q curve or where to search for pI.

        """
        return [self.pH_lower_bound, self.pH_upper_bound]

    def calculate_charged_curve(self, base_pkas, acid_pkas, constant_q=0):
        """Calculates the pH/Q curve."""
        for i in range(len(self.pH_range)):
            charge = PKaChargeCalculator().calculate_charge(
                base_pkas, acid_pkas, self.pH_range[i], constant_q=constant_q
            )
            self.q_range[i] = charge
        return np.vstack((self.pH_range, self.q_range)).T


class IsoelectricCalculator(object):
    """Calculates the isoelectric point and interval."""

    def __init__(self):
        self.tolerance = 0.01
        self.charge_tolerance = 0.05
        self.interval_threshold = 0.2
        self.pH_limit = CurveCalculator().get_pH_span()
        self.lower_pH = self.pH_limit[0]
        self.higher_pH = self.pH_limit[1]

    def calculate_pI(self, base_pkas, acid_pkas, constant_q=0):
        """
        Uses the pKas and charge to iteratively calculate
        the isoelectric point of a molecule.

        """
        while True:
            self.middle_pH = 0.5 * (self.higher_pH + self.lower_pH)
            charge = PKaChargeCalculator().calculate_charge(
                base_pkas, acid_pkas, self.middle_pH, constant_q=constant_q
            )
            na = len(acid_pkas)
            nb = len(base_pkas)

            if na == 0 and nb != 0:
                log.debug(
                    "Warning: no acidic ionizable groups, "
                    "only basic groups present in the "
                    "sequence. pI is not defined and thus "
                    "won't be calculated"
                    ""
                )
                reference_charge = self.charge_tolerance * nb
            elif nb == 0 and na != 0:
                log.debug(
                    "Warning: no basic ionizable groups, "
                    "only acidic groups present in the "
                    "sequence. pI is not defined and thus "
                    "won't be calculated"
                )
                reference_charge = -self.charge_tolerance * na
            else:
                reference_charge = 0.0

            # sic - instance attributes are updates as the
            # while block keeps looping
            if charge > reference_charge + self.tolerance:
                self.lower_pH = self.middle_pH
            elif charge < reference_charge - self.tolerance:
                self.higher_pH = self.middle_pH
            else:
                return self.middle_pH

            # If middle pH ends up outside the bounds
            # then return the bounds instead
            if self.middle_pH <= self.pH_limit[0]:
                return self.pH_limit[0]
            elif self.middle_pH >= self.pH_limit[1]:
                return self.pH_limit[1]

    def calculate_interval(self, pH_q):
        """Calculates the isoelectric interval."""
        q = pH_q[:, 1]
        pH = pH_q[:, 0]
        return pH[(q > -self.interval_threshold) & (q < self.interval_threshold)]



import tempfile

import pytest
from helpers import examples_dir
from pichemist.api import pichemist_from_dict
from pichemist.io import generate_input


def test_pka_matcher_json_1_file():
    """Example with mixed amino acids using pKaMatcher."""
    args = {
        "input_data": f"{examples_dir}/payload_1.smi",
        "input_format": "smiles_file",
        "plot_ph_q_curve": False,
        "print_fragments": False,
        "method": "pkamatcher",
    }

    expected = {
        1: {
            "mol_name": "Phe-Ornithine-aMeAsp-Lys-dAla",
            "pI": {
                "IPC2_peptide": 8.046875,
                "IPC_peptide": 9.8125,
                "ProMoST": 8.375,
                "Gauci": 8.6875,
                "Grimsley": 8.9375,
                "Thurlkill": 9.0625,
                "Lehninger": 9.859375,
                "Toseland": 9.40625,
                "pI mean": 9.0234375,
                "std": 1.721588565104915,
                "err": 0.6086734743994516,
            },
            "QpH7": {
                "IPC2_peptide": 0.6314906212267486,
                "IPC_peptide": 0.9915539516610472,
                "ProMoST": 0.26174063515548607,
                "Gauci": 0.5540630760817584,
                "Grimsley": 0.6645409545014482,
                "Thurlkill": 0.797542965316429,
                "Lehninger": 0.9932283675959863,
                "Toseland": 0.9515959465104951,
                "Q at pH7.4 mean": 0.7307195647561748,
                "std": 0.6749606913955383,
                "err": 0.23863464096007284,
            },
            "pI_interval": (8.624999999999998, 9.362499999999997),
            "pI_interval_threshold": 0.2,
            "pKa_set": "IPC2_peptide",
        }
    }

    input_dict = generate_input(args["input_format"], args["input_data"])
    output = pichemist_from_dict(
        input_dict, args["method"], args["plot_ph_q_curve"], args["print_fragments"]
    )
    assert expected == output


def test_pka_matcher_json_1_stdin():
    """Example with mixed amino acids using pKaMatcher."""
    args = {
        "input_data": "C[C@@H](NC(=O)[C@H](CCCCN)NC(=O)[C@](C)(CC(=O)O)NC(=O)[C@H](CCCN)NC(=O)[C@@H](N)Cc1ccccc1)C(=O)O",  # noqa
        "input_format": "smiles_stdin",
        "plot_ph_q_curve": False,
        "print_fragments": False,
        "method": "pkamatcher",
    }

    expected = {
        1: {
            "mol_name": "C[C@@H](NC(=O)[C@H](CCCCN)NC(=O)[C@](C)(CC(=O)O)NC(=O)[C@H](CCCN)NC(=O)[C@@H](N)Cc1ccccc1)C(=O)O",  # noqa
            "pI": {
                "IPC2_peptide": 8.046875,
                "IPC_peptide": 9.8125,
                "ProMoST": 8.375,
                "Gauci": 8.6875,
                "Grimsley": 8.9375,
                "Thurlkill": 9.0625,
                "Lehninger": 9.859375,
                "Toseland": 9.40625,
                "pI mean": 9.0234375,
                "std": 1.721588565104915,
                "err": 0.6086734743994516,
            },
            "QpH7": {
                "IPC2_peptide": 0.6314906212267486,
                "IPC_peptide": 0.9915539516610472,
                "ProMoST": 0.26174063515548607,
                "Gauci": 0.5540630760817584,
                "Grimsley": 0.6645409545014482,
                "Thurlkill": 0.797542965316429,
                "Lehninger": 0.9932283675959863,
                "Toseland": 0.9515959465104951,
                "Q at pH7.4 mean": 0.7307195647561748,
                "std": 0.6749606913955383,
                "err": 0.23863464096007284,
            },
            "pI_interval": (8.624999999999998, 9.362499999999997),
            "pI_interval_threshold": 0.2,
            "pKa_set": "IPC2_peptide",
        }
    }

    input_dict = generate_input(args["input_format"], args["input_data"])
    output = pichemist_from_dict(
        input_dict, args["method"], args["plot_ph_q_curve"], args["print_fragments"]
    )
    assert expected == output


def test_natural_aa_json_1():
    """Example with only natural amino acids (only FASTA matching)."""
    args = {
        "input_data": f"{examples_dir}/payload_2.smi",
        "input_format": "smiles_file",
        "plot_ph_q_curve": False,
        "print_fragments": False,
        "method": "pkamatcher",
    }

    expected = {
        1: {
            "mol_name": "Cys-Asn-Cys-Asn",
            "pI": {
                "IPC2_peptide": 5.0,
                "IPC_peptide": 5.0,
                "ProMoST": 5.5,
                "Gauci": 5.0,
                "Grimsley": 4.875,
                "Thurlkill": 5.5,
                "Lehninger": 5.0,
                "Toseland": 4.875,
                "pI mean": 5.09375,
                "std": 0.6789237807000135,
                "err": 0.240035804620894,
            },
            "QpH7": {
                "IPC2_peptide": -0.23913646012640216,
                "IPC_peptide": -0.23180587970079417,
                "ProMoST": -0.47710727633330485,
                "Gauci": -0.9370497315178979,
                "Grimsley": -1.9322611310762747,
                "Thurlkill": -0.33280195754105635,
                "Lehninger": -0.21536815651977892,
                "Toseland": -1.5908867442915533,
                "Q at pH7.4 mean": -0.7445521671383828,
                "std": 1.7898169252151255,
                "err": 0.6327958424510355,
            },
            "pI_interval": (3.7624999999999984, 6.687499999999999),
            "pI_interval_threshold": 0.2,
            "pKa_set": "IPC2_peptide",
        }
    }

    input_dict = generate_input(args["input_format"], args["input_data"])
    output = pichemist_from_dict(
        input_dict, args["method"], args["plot_ph_q_curve"], args["print_fragments"]
    )
    assert expected == output


@pytest.mark.acd
def test_acd_json_1():
    """Example with mixed amino acids using ACD."""
    args = {
        "input_data": f"{examples_dir}/payload_1.smi",
        "input_format": "smiles_file",
        "plot_ph_q_curve": False,
        "print_fragments": False,
        "method": "acd",
    }

    expected = {
        1: {
            "mol_name": "Phe-Ornithine-aMeAsp-Lys-dAla",
            "pI": {
                "IPC2_peptide": 8.046875,
                "IPC_peptide": 9.75,
                "ProMoST": 8.3125,
                "Gauci": 8.625,
                "Grimsley": 8.875,
                "Thurlkill": 9.0,
                "Lehninger": 9.796875,
                "Toseland": 9.34375,
                "pI mean": 8.96875,
                "std": 1.686776465258512,
                "err": 0.5963655384650843,
            },
            "QpH7": {
                "IPC2_peptide": 0.6313327301392769,
                "IPC_peptide": 0.9913960605735755,
                "ProMoST": 0.2615827440680144,
                "Gauci": 0.5539051849942872,
                "Grimsley": 0.6643830634139766,
                "Thurlkill": 0.7973850742289573,
                "Lehninger": 0.9930704765085147,
                "Toseland": 0.9514380554230234,
                "Q at pH7.4 mean": 0.7305616736687033,
                "std": 0.6749606913955383,
                "err": 0.23863464096007284,
            },
            "pI_interval": (8.612499999999997, 9.312499999999996),
            "pI_interval_threshold": 0.2,
            "pKa_set": "IPC2_peptide",
        }
    }

    input_dict = generate_input(args["input_format"], args["input_data"])
    output = pichemist_from_dict(
        input_dict, args["method"], args["plot_ph_q_curve"], args["print_fragments"]
    )
    assert expected == output


@pytest.mark.acd
def test_acd_json_2():
    """Example with mixed amino acids using ACD (GALAS)."""
    args = {
        "input_data": f"{examples_dir}/payload_6.smi",
        "input_format": "smiles_file",
        "plot_ph_q_curve": False,
        "print_fragments": False,
        "method": "acd",
    }

    expected = {
        1: {
            "mol_name": "Galas789",
            "pI": {
                "IPC2_peptide": 7.875,
                "IPC_peptide": 7.75,
                "ProMoST": 8.125,
                "Gauci": 7.75,
                "Grimsley": 8.0,
                "Thurlkill": 8.0,
                "Lehninger": 7.75,
                "Toseland": 7.875,
                "pI mean": 7.890625,
                "std": 0.37238672774415577,
                "err": 0.13165859020588058,
            },
            "QpH7": {
                "IPC2_peptide": 0.08874508270557702,
                "IPC_peptide": 0.029950764546196496,
                "ProMoST": 0.22623319165452238,
                "Gauci": 0.02686668488169408,
                "Grimsley": 0.12698532501860005,
                "Thurlkill": 0.11158023273557527,
                "Lehninger": 0.028394223343287228,
                "Toseland": 0.06859834054253844,
                "Q at pH7.4 mean": 0.08841923067849887,
                "std": 0.1794985133244505,
                "err": 0.0634623079923114,
            },
            "pI_interval": (6.987499999999999, 8.799999999999997),
            "pI_interval_threshold": 0.2,
            "pKa_set": "IPC2_peptide",
        }
    }

    input_dict = generate_input(args["input_format"], args["input_data"])
    output = pichemist_from_dict(
        input_dict, args["method"], args["plot_ph_q_curve"], args["print_fragments"]
    )
    assert expected == output


@pytest.mark.acd
def test_natural_aa_json_2():
    """Example with only natural amino acids (only FASTA matching)."""
    args = {
        "input_data": f"{examples_dir}/payload_2.smi",
        "input_format": "smiles_file",
        "plot_ph_q_curve": False,
        "print_fragments": False,
        "method": "acd",
    }

    expected = {
        1: {
            "mol_name": "Cys-Asn-Cys-Asn",
            "pI": {
                "IPC2_peptide": 5,
                "IPC_peptide": 5,
                "ProMoST": 5.5,
                "Gauci": 5,
                "Grimsley": 4.875,
                "Thurlkill": 5.5,
                "Lehninger": 5,
                "Toseland": 4.875,
                "pI mean": 5.09375,
                "std": 0.6789237807000135,
                "err": 0.240035804620894,
            },
            "QpH7": {
                "IPC2_peptide": -0.23913646012640216,
                "IPC_peptide": -0.23180587970079417,
                "ProMoST": -0.47710727633330485,
                "Gauci": -0.9370497315178979,
                "Grimsley": -1.9322611310762747,
                "Thurlkill": -0.33280195754105635,
                "Lehninger": -0.21536815651977892,
                "Toseland": -1.5908867442915533,
                "Q at pH7.4 mean": -0.7445521671383828,
                "std": 1.7898169252151255,
                "err": 0.6327958424510355,
            },
            "pI_interval": (3.7624999999999984, 6.687499999999999),
            "pI_interval_threshold": 0.2,
            "pKa_set": "IPC2_peptide",
        }
    }

    input_dict = generate_input(args["input_format"], args["input_data"])
    output = pichemist_from_dict(
        input_dict, args["method"], args["plot_ph_q_curve"], args["print_fragments"]
    )
    assert expected == output


def test_natural_aa_json_3():
    """
    Example with only natural amino acids (only FASTA matching).
    It also checks for the presence of "plot_filename" in the output.

    """
    tmp_file_prefix = tempfile.NamedTemporaryFile().name
    tmp_filepath = f"{tmp_file_prefix}_1.png"
    args = {
        "input_data": f"{examples_dir}/payload_2.smi",
        "input_format": "smiles_file",
        "ph_q_curve_file_prefix": tmp_file_prefix,
        "plot_ph_q_curve": True,
        "print_fragments": False,
        "method": "pkamatcher",
    }

    expected = {
        1: {
            "mol_name": "Cys-Asn-Cys-Asn",
            "pI": {
                "IPC2_peptide": 5.0,
                "IPC_peptide": 5.0,
                "ProMoST": 5.5,
                "Gauci": 5.0,
                "Grimsley": 4.875,
                "Thurlkill": 5.5,
                "Lehninger": 5.0,
                "Toseland": 4.875,
                "pI mean": 5.09375,
                "std": 0.6789237807000135,
                "err": 0.240035804620894,
            },
            "QpH7": {
                "IPC2_peptide": -0.23913646012640216,
                "IPC_peptide": -0.23180587970079417,
                "ProMoST": -0.47710727633330485,
                "Gauci": -0.9370497315178979,
                "Grimsley": -1.9322611310762747,
                "Thurlkill": -0.33280195754105635,
                "Lehninger": -0.21536815651977892,
                "Toseland": -1.5908867442915533,
                "Q at pH7.4 mean": -0.7445521671383828,
                "std": 1.7898169252151255,
                "err": 0.6327958424510355,
            },
            "pI_interval": (3.7624999999999984, 6.687499999999999),
            "pI_interval_threshold": 0.2,
            "plot_filename": tmp_filepath,
            "pKa_set": "IPC2_peptide",
        }
    }

    input_dict = generate_input(args["input_format"], args["input_data"])
    output = pichemist_from_dict(
        input_dict,
        args["method"],
        args["ph_q_curve_file_prefix"],
        args["plot_ph_q_curve"],
        args["print_fragments"],
    )
    assert expected == output



from pichemist.molecule import MolStandardiser
from rdkit import Chem
from rdkit import RDLogger

RDLogger.DisableLog("rdApp.info")


class SmartsChargeCalculator(object):
    """Calculates the nitrogen cation charges for an input using SMARTS."""

    def __init__(self):
        self.smarts = {
            # Positively charged N but not connected to any negatively
            # charged atom. To avoid azido and nitro groups being counted
            "nitrogen_cation": "[#7+;!H1;!H2;!H3;!H4!$([#7+]~[*-])]"
        }
        self.patterns = self._prepare_patterns()

    def _prepare_patterns(self):
        """Converts SMARTS strings into objects."""
        return {name: Chem.MolFromSmarts(s) for name, s in self.smarts.items()}

    def _get_mol_from_smiles(self, smiles):
        """Get mol object from SMILES."""
        return Chem.MolFromSmiles(smiles)

    def calculate_net_qs_from_list(self, smiles_list):
        """Produces a list of charges and matching SMILES."""
        net_qs = list()
        for smiles in smiles_list:
            mol = self._get_mol_from_smiles(smiles)
            mol = MolStandardiser().standardise_molecule(mol)
            # sic - Appends charge 1 and the SMILES for each match
            for _ in self._get_net_qs_matches_from_mol(mol):
                net_qs.append((1, smiles))
        return net_qs

    def _get_net_qs_matches_from_mol(self, mol):
        """
        Matches a nitrogen cation pattern against a
        molecule and returns a list of matches.

        """
        pattern = self.patterns["nitrogen_cation"]
        matches = mol.GetSubstructMatches(pattern)
        return [y[0] for y in matches]

    def calculate_net_qs_from_smiles(self, smiles):
        """Returns the number of charges for a given SMILES."""
        mol = self._get_mol_from_smiles(smiles)
        mol = MolStandardiser().standardise_molecule(mol)
        # sic - The charges corresponds to the number of matches
        number_of_charges = len(self._get_net_qs_matches_from_mol(mol))
        return number_of_charges


class PKaChargeCalculator(object):
    """Calculate the molecule charge given its pKas."""

    def __init__(self):
        pass

    def _calculate_basic_charge(self, pH, pKa):
        """Calculate charge contribution by basic pKas."""
        return 1 / (1 + 10 ** (pH - pKa))

    def _calculate_acidic_charge(self, pH, pKa):
        """Calculate charge contribution by acidic pKas."""
        return -1 / (1 + 10 ** (pKa - pH))

    def _calculate_diacidic_charge(self, pH, pKa1, pKa2):
        """Calculate charge contribution by diacidic pKas."""
        Ka1 = 10 ** (-pKa1)
        Ka2 = 10 ** (-pKa2)
        H = 10 ** (-pH)
        f1 = (H * Ka1) / (H**2 + H * Ka1 + Ka1 * Ka2)  # fraction of [AH-]
        f2 = f1 * Ka2 / H  # fraction of [A2-]
        return -2 * f2 + (-1) * f1  # average charge of phosphate

    def calculate_charge(self, base_pkas, acid_pkas, pH, constant_q=0):
        """Calculate the molecule charge from all contributions."""
        charge = constant_q
        for pka in base_pkas:
            charge += self._calculate_basic_charge(pH, pka)
        for pka in acid_pkas:
            charge += self._calculate_acidic_charge(pH, pka)
        return charge

    def calculate_constant_charge(self, net_qs):
        """Calculates the constant charge from the net charges."""
        constant_q = 0.0
        if len(net_qs) > 0:
            constant_q = float(sum(net_qs)) / float(len(net_qs))
        return constant_q



import json
import os
import tempfile

import pytest
from helpers import examples_dir
from helpers import stdout_to_variable
from helpers import TestError
from pichemist.cli import arg_parser
from pichemist.cli import run_pichemist


def test_parser_creation():
    """Parser creation."""
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_1.smi",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    assert args is not None


def test_console_text_output_1():
    """Validity of console text output using pKaMatcher."""
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_1.smi",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    with open(f"{examples_dir}/payload_1_out.txt", "r") as f:
        expected = f.read()
    assert result == expected


def test_console_text_output_2():
    """Validity of console text output using pKaMatcher."""
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_3.smi",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    with open(f"{examples_dir}/payload_3_out.txt", "r") as f:
        expected = f.read()
    assert result == expected


@pytest.mark.acd
def test_console_text_output_3():
    """Validity of console text output using ACD."""
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_5.smi",
            "--print_fragment_pkas",
            "--method",
            "acd",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    with open(f"{examples_dir}/payload_5_out.txt", "r") as f:
        expected = f.read()
    assert result == expected


def test_console_json_output():
    """Validity of console JSON output."""
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_1.smi",
            "-of",
            "json",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    result = json.loads(result)
    with open(f"{examples_dir}/payload_1_out.json", "r") as f:
        expected = json.load(f)
    assert result == expected


def test_console_json_output_images():
    """Validity of console JSON output."""
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_1.smi",
            "-of",
            "json",
            "--print_fragment_pkas",
            "--generate_fragment_images",
            "--method",
            "pkamatcher",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    result = json.loads(result)
    with open(f"{examples_dir}/payload_1_out_images.json", "r") as f:
        expected = json.load(f)
    assert result == expected


def test_file_csv_output():
    """Validity of CSV file output."""
    tmp_filepath = tempfile.NamedTemporaryFile(suffix=".csv").name
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_1.smi",
            "-o",
            tmp_filepath,
            "-of",
            "csv",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    run_pichemist(args)
    if not os.path.exists(tmp_filepath):
        raise TestError("File was not created.")
    with open(f"{examples_dir}/payload_1_out.csv", "r") as f:
        expected = f.read()
        with open(tmp_filepath, "r") as f:
            results = f.read()
            assert results == expected
    os.remove(tmp_filepath)


def test_file_sdf_output_1():
    """Validity of SDF file output."""
    tmp_filepath = tempfile.NamedTemporaryFile(suffix=".sdf").name
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_1.smi",
            "-o",
            tmp_filepath,
            "-of",
            "sdf",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    run_pichemist(args)
    if not os.path.exists(tmp_filepath):
        raise TestError("File was not created.")
    with open(f"{examples_dir}/payload_1_out.sdf", "r") as f:
        expected = f.read()
        with open(tmp_filepath, "r") as f:
            results = f.read()
            assert results == expected
    os.remove(tmp_filepath)


def test_file_sdf_output_2():
    """Validity of SDF file output."""
    tmp_filepath = tempfile.NamedTemporaryFile(suffix=".sdf").name
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_3.smi",
            "-o",
            tmp_filepath,
            "-of",
            "sdf",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    run_pichemist(args)
    if not os.path.exists(tmp_filepath):
        raise TestError("File was not created.")
    with open(f"{examples_dir}/payload_3_out.sdf", "r") as f:
        expected = f.read()
        with open(tmp_filepath, "r") as f:
            results = f.read()
            assert results == expected
    os.remove(tmp_filepath)


def test_file_sdf_input_output():
    """Validity of SDF file input and output."""
    tmp_filepath = tempfile.NamedTemporaryFile(suffix=".sdf").name
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_4.sdf",
            "-if",
            "sdf",
            "-o",
            tmp_filepath,
            "-of",
            "sdf",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    run_pichemist(args)
    if not os.path.exists(tmp_filepath):
        raise TestError("File was not created.")
    with open(f"{examples_dir}/payload_4_out.sdf", "r") as f:
        expected = f.read()
        with open(tmp_filepath, "r") as f:
            results = f.read()
            assert results == expected
    os.remove(tmp_filepath)


def test_smiles_stdin_input_1():
    """Validity of SMILES stdin input and text output."""
    args = arg_parser(
        [
            "-i",
            "C[C@@H](NC(=O)[C@H](CCCCN)NC(=O)[C@](C)"
            "(CC(=O)O)NC(=O)[C@H](CCCN)NC(=O)[C@@H]"
            "(N)Cc1ccccc1)C(=O)O",
            "-if",
            "smiles_stdin",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    with open(f"{examples_dir}/payload_1_out.txt", "r") as f:
        expected = f.read()
    assert result == expected


def test_smiles_stdin_input_2():
    """Validity of SMILES stdin input and text output."""
    args = arg_parser(
        [
            "-i",
            "N[C@@]([H])(CS)C(=O)N[C@@]([H])(CC(=O)N)C(=O)N[C@@]([H])(CS)C(=O)N[C@@]([H])(CC(=O)N)C(=O)O",  # noqa: E501
            "-if",
            "smiles_stdin",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    with open(f"{examples_dir}/payload_2_out_smiles.txt", "r") as f:
        expected = f.read()
    assert result == expected


def test_fasta_stdin_input():
    """Validity of FASTA stdin input and text output."""
    args = arg_parser(
        [
            "-i",
            "CNCN",
            "-if",
            "fasta_stdin",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    with open(f"{examples_dir}/payload_2_out_fasta.txt", "r") as f:
        expected = f.read()
    assert result == expected


def test_fasta_stdin_capped_cterm_input():
    """Validity of FASTA stdin input and text output."""
    args = arg_parser(
        [
            "-i",
            "CNCN",
            "-if",
            "fasta_stdin",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
            "--ionizable_cterm",
            "false",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    with open(f"{examples_dir}/payload_2_out_fasta_capped_cterm.txt", "r") as f:
        expected = f.read()
    assert result == expected


def test_fasta_stdin_capped_nterm_input():
    """Validity of FASTA stdin input and text output."""
    args = arg_parser(
        [
            "-i",
            "CNCN",
            "-if",
            "fasta_stdin",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
            "--ionizable_nterm",
            "false",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    with open(f"{examples_dir}/payload_2_out_fasta_capped_nterm.txt", "r") as f:
        expected = f.read()
    assert result == expected


def test_fasta_stdin_all_capped_input():
    """Validity of FASTA stdin input and text output."""
    args = arg_parser(
        [
            "-i",
            "CNCN",
            "-if",
            "fasta_stdin",
            "--print_fragment_pkas",
            "--method",
            "pkamatcher",
            "--ionizable_nterm",
            "false",
            "--ionizable_cterm",
            "false",
        ]
    )
    result = stdout_to_variable(run_pichemist, args)
    with open(f"{examples_dir}/payload_2_out_fasta_all_capped.txt", "r") as f:
        expected = f.read()
    assert result == expected


def test_file_ph_q_plot_1():
    """Existence of the pH/Q plot file."""
    tmp_file_prefix = tempfile.NamedTemporaryFile().name
    tmp_filepath = f"{tmp_file_prefix}_1.png"
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_1.smi",
            "-of",
            "json",
            "--print_fragment_pkas",
            "-pp",
            tmp_file_prefix,
            "--plot_ph_q_curve",
            "--method",
            "pkamatcher",
        ]
    )
    _ = stdout_to_variable(run_pichemist, args)
    if not os.path.exists(tmp_filepath):
        raise TestError("File was not created.")
    os.remove(tmp_filepath)


def test_file_ph_q_plot_2():
    """Existence of the pH/Q plot file."""
    tmp_file_prefix = tempfile.NamedTemporaryFile().name
    tmp_filepath = f"{tmp_file_prefix}_1.png"
    args = arg_parser(
        [
            "-i",
            f"{examples_dir}/payload_2.smi",
            "-of",
            "json",
            "--print_fragment_pkas",
            "-pp",
            tmp_file_prefix,
            "--plot_ph_q_curve",
            "--method",
            "pkamatcher",
        ]
    )
    _ = stdout_to_variable(run_pichemist, args)
    if not os.path.exists(tmp_filepath):
        raise TestError("File was not created.")
    os.remove(tmp_filepath)



import math


def mean(input_list):
    """Calculates the mean for an input list."""
    return sum(input_list) / len(input_list)


def stddev(input_list):
    """Calculates the standard deviation for an input list."""
    mn = mean(input_list)
    variance = sum([(e-mn)**2 for e in input_list])
    return math.sqrt(variance)


def stderr(input_list):
    """Calculates the standard error for an input list."""
    mn = mean(input_list)
    variance = sum([(e-mn)**2 for e in input_list])
    return math.sqrt(variance) / math.sqrt(len(input_list))



from pichemist.config import AA_TYPE_KEYS
from pichemist.fasta.helpers import pattern_match
from pichemist.fasta.smarts import AA_SMARTS_SET


class FastaScrambler(object):
    def __init__(self):
        self.aa_smarts_set = AA_SMARTS_SET

    def get_scrambled_fasta_from_list(self, smiles_list):
        """
        Retrieves amino acid letters (FASTA) for a list of
        SMILES and concatenates them into a string.

        """
        aa_list = list()
        for smiles in smiles_list:
            match = False

            # Middle
            CAPPED_KEY = AA_TYPE_KEYS["capped"]
            CAPPED_SMARTS = self.aa_smarts_set[CAPPED_KEY]
            for smarts, aa in CAPPED_SMARTS.items():
                nhits = pattern_match(smiles, smarts)
                if nhits > 0:
                    aa_list += [aa]*nhits
                    match = True
                    break

            # N-term
            NTERM_KEY = AA_TYPE_KEYS["nterm"]
            NTERM_SMARTS = self.aa_smarts_set[NTERM_KEY]
            for smarts, aa in NTERM_SMARTS.items():
                nhits = pattern_match(smiles, smarts)
                if nhits > 0:
                    aa_list += [aa]*nhits
                    match = True
                    break

            # C-term
            CTERM_KEY = AA_TYPE_KEYS["cterm"]
            CTERM_SMARTS = self.aa_smarts_set[CTERM_KEY]
            for smarts, aa in CTERM_SMARTS.items():
                nhits = pattern_match(smiles, smarts)
                if nhits > 0:
                    aa_list += [aa]*nhits
                    match = True
                    break

            if not match:
                match = False
                aa_list += ["X"]
        return "".join(aa_list)



import pichemist

from pichemist.model import ACDPKaFlag

# FASTA definitions
KNOWN_BASIC_RESIDUES = ["K", "R", "H"]
KNOWN_ACIDIC_RESIDUES = ["D", "E", "C", "Y", "U"]
KNOWN_RESIDUES = ["G", "A", "S", "P", "V", "T", "C",
                  "I", "L", "N", "D", "Q", "K", "E",
                  "M", "H", "F", "R", "Y", "W", "X",
                  "Z", "B", "U"]

PKA_JSON_INDICES = {
     "primary": 0,
     "n-terminal": 1,
     "c-terminal": 2
}

REFERENCE_PKA_SET = "IPC2_peptide"

PKA_SETS_NAMES = ["IPC2_peptide",
                  "IPC_peptide",
                  "ProMoST",
                  "Gauci",
                  "Grimsley",
                  "Thurlkill",
                  "Lehninger",
                  "Toseland"]

ALL_KNOWN_PKA_SETS_NAMES = ["ProMoST",
                            "IPC_peptide",
                            "IPC2_peptide",
                            "Gauci",
                            "Bjellqvist",
                            "Rodwell",
                            "Grimsley",
                            "Thurlkill",
                            "EMBOSS",
                            "DTASelect",
                            "Solomon",
                            "Sillero",
                            "Lehninger",
                            "Toseland",
                            "Nozaki",
                            "Dawson"]

# SMARTS definitions
SKIP_SMARTS_NAMES = []
PKA_LIMITS = {
    "acid_1": -5,
    "acid_2": 12,
    "base_1": 2,
    "base_2": 15,
}
AA_TYPE_KEYS = {
    "capped": "d_capped_aa_smarts",
    "nterm": "d_nterm_free_aa_smarts",
    "cterm": "d_cterm_free_aa_smarts"
}

# ACD configuration
ACD_METHOD = ACDPKaFlag.GALAS

# Plot definitions
PLOT_LINE_WIDTHS = {
    "w1": 4.0,
    "w2": 3.0,
    "w3": 2.0,
    "w4": 1.0
}

# Paths
MODULE_DIR = pichemist.__path__[0]
DATA_DIR = f"{MODULE_DIR}/data"
FASTA_PKA_DATA_DIR = f"{DATA_DIR}/fasta"
FASTA_PKA_SETS_DIR = f"{FASTA_PKA_DATA_DIR}/standardised"
SMARTS_PKA_DATA_DIR = f"{DATA_DIR}/smarts"
SS_SMARTS_PKA_SET_FILEPATH = f"{SMARTS_PKA_DATA_DIR}/standardised/" \
                            "ss_smarts_set_standardised.json"
AA_SMARTS_SET_FILEPATH = f"{SMARTS_PKA_DATA_DIR}/standardised/" \
                         "aa_smarts_set_standardised.json"
PH_Q_FILE_PREFIX = "out_ph_q_curve"



notes = {
        "2022-07-15": "compressed imidazole smarts; split amines into primary, secondary, tertiary; change carboxylic acid value to ACDlab prediction; added monoand di-halogenated phenols",
        "2022-07-11": "added smarts patters to match tetrazole, substituted phenol, substituted imidazoles, pyridines. Use pKa values predicted by ACDlabs GALAS method; fixed most of smarts patters such that they explicit hydrogens compressed to heavy atom definition, otherwise match did not work; removed very generic definitions, like, aromatic N deprotonated as they hit too many unwanted structures",
        "2020-09-11": "added last column with acid-base classification"
    }


import inspect
import sys
from enum import Enum


class BaseEnum(str, Enum):
    """Base class for models."""

    pass


class PKaType(BaseEnum):
    ACIDIC = "acid"
    BASIC = "base"


class PKaMethod(BaseEnum):
    PKA_MATCHER = "pkamatcher"
    ACD = "acd"


class ACDPKaFlag(BaseEnum):
    CLASSIC = "-MPKAAPP"
    GALAS = "-MPKAAPPGALAS"


class InputAttribute(BaseEnum):
    MOL_NAME = "mol_name"
    MOL_OBJECT = "mol_obj"
    MOL_FASTA = "fasta"


class InputFormat(BaseEnum):
    SMILES_STDIN = "smiles_stdin"
    SMILES_FILE = "smiles_file"
    FASTA_STDIN = "fasta_stdin"
    FASTA_FILE = "fasta_file"
    SD_FILE = "sdf"


class InputFileExtension(BaseEnum):
    SDF = "sdf"
    SMILES = "smi"
    FASTA = "fasta"


class OutputAttribute(BaseEnum):
    MOL_NAME = "mol_name"
    SMILES = "SMILES"
    PI = "pI"
    Q_PH7 = "QpH7"
    PI_INTERVAL = "pI_interval"
    PI_INTERVAL_THRESHOLD = "pI_interval_threshold"
    PLOT_FILENAME = "plot_filename"
    PKA_SET = "pKa_set"
    FRAG_BASE_PKA_FASTA = "frag_base_pkas_fasta"
    FRAG_ACID_PKA_FASTA = "frag_acid_pkas_fasta"
    FRAG_BASE_PKA_CALC = "frag_base_pkas_calc"
    FRAG_ACID_PKA_CALC = "frag_acid_pkas_calc"
    FRAG_CONSTANT_QS = "frag_Qs_calc"


class OutputFragAttribute(BaseEnum):
    TYPE = "type"
    COUNT = "count"
    PKA = "pka"
    FRAGMENT = "frag"
    IMAGE = "base64_image"


class OutputFormat(BaseEnum):
    JSON = "json"
    SD_FILE = "sdf"
    CSV_FILE = "csv"
    CONSOLE = "console"


class OutputConsoleFormat(BaseEnum):
    JSON = "json"
    CONSOLE = "console"


class OutputFileFormat(BaseEnum):
    SD_FILE = "sdf"
    CSV_FILE = "csv"


class Models(object):
    """Contains the definitions of all data models."""

    def __init__(self):
        self.definitions = self._get_definitions()

    @staticmethod
    def _get_subclasses_of_enum():
        """Produces a dict of names and subclasses of BaseEnum."""
        classes = dict()
        for name, cls in inspect.getmembers(sys.modules[__name__]):
            if inspect.isclass(cls) and cls is not BaseEnum:
                if issubclass(cls, BaseEnum):
                    classes[name] = cls
        return classes

    @staticmethod
    def _get_values_from_enum(BaseEnum):
        """Generates a list from all values of a class."""
        return [BaseEnum[x].value for x in BaseEnum._member_names_]

    def _get_definitions(self):
        """Builds a dictionary with classes and their values."""
        classes = self._get_subclasses_of_enum()
        return {c: self._get_values_from_enum(c) for c in classes.values()}


MODELS = Models().definitions



from pichemist.config import AA_TYPE_KEYS
from pichemist.config import PKA_SETS_NAMES
from pichemist.fasta.helpers import pattern_match
from pichemist.fasta.pka_sets import FASTA_PKA_SETS
from pichemist.fasta.smarts import AA_SMARTS_SET


class FastaPKaMatcher(object):
    def __init__(self):
        self.pka_sets = PKA_SETS_NAMES
        self.aa_smarts_set = AA_SMARTS_SET

    def _initialise_pka_sets(
        self, all_pka_sets=FASTA_PKA_SETS, pka_names=PKA_SETS_NAMES
    ):
        """Generates a filtered set of pKa sets."""
        pka_sets = dict()
        for n in pka_names:
            pka_sets[n] = all_pka_sets[n]
        return pka_sets

    def get_pka_sets_names(self):
        """Getter for the names of the sets."""
        return self.pka_sets

    def _initialise_pka_dicts(self, pka_sets):
        """Initialises empty pKa dictionaries."""
        base_pka_dict = dict()
        acid_pka_dict = dict()
        diacid_pka_dict = dict()

        for n in pka_sets:
            base_pka_dict[n] = list()
            acid_pka_dict[n] = list()
            diacid_pka_dict[n] = list()
        return base_pka_dict, acid_pka_dict, diacid_pka_dict

    def _get_pka_from_set(self, aa, pka_set, pka_idx):
        """Returns a pKa value from a set."""
        return pka_set[aa][pka_idx]

    def _add_pka_to_dict(self, aa, name, pka, pka_dict):
        """Assigns a (pKa, AA) tuple to a dictionary key."""
        pka_dict[name].append((pka, aa))
        return pka_dict

    def _add_pka_to_dict_if_present(self, aa, name, pka_dict, pka_set, pka_idx):
        """Adds a pKa value to a dict if present in a query set."""
        if aa in pka_set.keys():
            pka = self._get_pka_from_set(aa, pka_set, pka_idx)
            pka_dict = self._add_pka_to_dict(aa, name, pka, pka_dict)

    def _add_pka_to_acidic_and_basic_dicts(
        self, aa, pka_set_name, base_pka_dict, acid_pka_dict, pka_set, aa_idx
    ):
        """
        Wraps the logic to add a pKa value to
        acidic and basic dicts.

        """
        self._add_pka_to_dict_if_present(
            aa, pka_set_name, base_pka_dict, pka_set["basic"], aa_idx
        )
        self._add_pka_to_dict_if_present(
            aa, pka_set_name, acid_pka_dict, pka_set["acidic"], aa_idx
        )

    def _build_name(self, aa, aa_type):
        """Builds a name with AA and its type."""
        return f"{aa}_{aa_type}"

    def _add_terminus_ionizable_to_dict(
        self, aa, pka_set_name, pka_dict, pka_set, aa_type, aa_idx
    ):
        """Adds a terminus pka value to a dict."""
        self._add_pka_to_dict(
            self._build_name(aa, aa_type),
            pka_set_name,
            self._get_pka_from_set(aa, pka_set["terminus_ionizable"], aa_idx),
            pka_dict,
        )

    def _get_aa_pkas(
        self,
        smiles,
        unknown_fragments,
        base_pka_dict,
        acid_pka_dict,
        diacid_pka_dict,  # sic - not used
        pka_sets,
    ):
        """
        For a given SMILES, it matches its pKa
        value against a set of pKa sets.
        NOTE: Tried to break this function down
        further but integration tests break.

        """
        match = False

        # Middle
        CAPPED_IDX = 0
        CAPPED_KEY = AA_TYPE_KEYS["capped"]
        CAPPED_SMARTS = self.aa_smarts_set[CAPPED_KEY]
        for smarts, aa in CAPPED_SMARTS.items():
            nhits = pattern_match(smiles, smarts)
            if nhits > 0:
                for n, pka_set in pka_sets.items():
                    # Add pKa values for each match
                    for _ in range(nhits):
                        self._add_pka_to_acidic_and_basic_dicts(
                            aa, n, base_pka_dict, acid_pka_dict, pka_set, CAPPED_IDX
                        )
                match = True
                break

        # N-term
        NTERM_FREE_IDX = 1
        NTERM_ION_IDX = 0
        NTERM_KEY = AA_TYPE_KEYS["nterm"]
        NTERM_SMARTS = self.aa_smarts_set[NTERM_KEY]
        for smarts, aa in NTERM_SMARTS.items():
            nhits = pattern_match(smiles, smarts)
            if nhits > 0:
                for n, pka_set in pka_sets.items():
                    # Add pKa values for each match
                    for _ in range(nhits):
                        self._add_pka_to_acidic_and_basic_dicts(
                            aa, n, base_pka_dict, acid_pka_dict, pka_set, NTERM_FREE_IDX
                        )
                        self._add_terminus_ionizable_to_dict(
                            aa, n, base_pka_dict, pka_set, "N-term", NTERM_ION_IDX
                        )
                match = True
                break

        # C-term
        CTERM_FREE_IDX = 2
        CTERM_ION_IDX = 1
        CTERM_KEY = AA_TYPE_KEYS["cterm"]
        CTERM_SMARTS = self.aa_smarts_set[CTERM_KEY]
        for smarts, aa in CTERM_SMARTS.items():
            nhits = pattern_match(smiles, smarts)
            if nhits > 0:
                for n, pka_set in pka_sets.items():
                    # Add pKa values for each match
                    for _ in range(nhits):
                        self._add_pka_to_acidic_and_basic_dicts(
                            aa, n, base_pka_dict, acid_pka_dict, pka_set, CTERM_FREE_IDX
                        )
                        self._add_terminus_ionizable_to_dict(
                            aa, n, acid_pka_dict, pka_set, "C-term", CTERM_ION_IDX
                        )
                match = True
                break

        # Append as unknown if not matched
        if not match:
            unknown_fragments.append(smiles)
        return unknown_fragments, base_pka_dict, acid_pka_dict, diacid_pka_dict

    def _get_aa_pkas_from_single_letter_aa(
        self,
        aa,
        base_pka_dict,
        acid_pka_dict,
        diacid_pka_dict,  # sic - not used
        pka_sets,
        aa_position_in_sequence,
    ):
        """
        For a given amino acids in single letter format,
        it matches its pKa
        value against a set of pKa sets.
        """

        # Middle
        if aa_position_in_sequence == "middle":
            CAPPED_IDX = 0
            for n, pka_set in pka_sets.items():
                self._add_pka_to_acidic_and_basic_dicts(
                    aa, n, base_pka_dict, acid_pka_dict, pka_set, CAPPED_IDX
                )

        # N-term, sidechains
        elif aa_position_in_sequence == "nterm sidechains":
            NTERM_FREE_IDX = 1
            for n, pka_set in pka_sets.items():
                self._add_pka_to_acidic_and_basic_dicts(
                    aa, n, base_pka_dict, acid_pka_dict, pka_set, NTERM_FREE_IDX
                )

        # N-term, nterm
        elif aa_position_in_sequence == "nterm nterm":
            NTERM_ION_IDX = 0
            for n, pka_set in pka_sets.items():
                self._add_terminus_ionizable_to_dict(
                    aa, n, base_pka_dict, pka_set, "N-term", NTERM_ION_IDX
                )

        # C-term, sidechains
        elif aa_position_in_sequence == "cterm sidechains":
            CTERM_FREE_IDX = 2
            for n, pka_set in pka_sets.items():
                self._add_pka_to_acidic_and_basic_dicts(
                    aa, n, base_pka_dict, acid_pka_dict, pka_set, CTERM_FREE_IDX
                )

        # C-term, cterm
        elif aa_position_in_sequence == "cterm cterm":
            CTERM_ION_IDX = 1
            for n, pka_set in pka_sets.items():
                self._add_terminus_ionizable_to_dict(
                    aa, n, acid_pka_dict, pka_set, "C-term", CTERM_ION_IDX
                )

        else:
            raise ValueError(
                f"Invalid `aa_position_in_sequence` (got {aa_position_in_sequence})"
            )

        return base_pka_dict, acid_pka_dict, diacid_pka_dict

    def get_aa_pkas_from_list(self, smiles_list):
        """
        Matches a list of SMILES against
        the pKa values in a set of pKa sets.

        """
        # Initialise pKa sets
        pka_sets = self._initialise_pka_sets()

        # Initialise results
        unknown_fragments = list()
        base_pka_dict, acid_pka_dict, diacid_pka_dict = self._initialise_pka_dicts(
            PKA_SETS_NAMES
        )

        for s in smiles_list:
            (
                unknown_fragments,
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
            ) = self._get_aa_pkas(
                s,
                unknown_fragments,
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
                pka_sets,
            )
        return unknown_fragments, base_pka_dict, acid_pka_dict, diacid_pka_dict

    def get_aa_pkas_from_aa_list(self, aa_list, ionizable_nterm, ionizable_cterm):
        """
        Matches a list of sigle-letter aminoa acids against
        the pKa values in a set of pKa sets.

        """
        # Initialise pKa sets
        pka_sets = self._initialise_pka_sets()

        # Initialise results
        base_pka_dict, acid_pka_dict, diacid_pka_dict = self._initialise_pka_dicts(
            PKA_SETS_NAMES
        )

        # ionizable both N- and C-termini
        if ionizable_cterm and ionizable_nterm:
            aa_middle_list = aa_list[1:-1]
            aa_nterm_ionizable_sidechain_list = [aa_list[0]]
            aa_nterm_ionizable_nterm_list = [aa_list[0]]
            aa_cterm_ionizable_sidechain_list = [aa_list[-1]]
            aa_cterm_ionizable_cterm_list = [aa_list[-1]]

        # capped C-terminus, sidechain still ionizable
        elif not ionizable_cterm and ionizable_nterm:
            aa_middle_list = aa_list[1:-1]
            aa_nterm_ionizable_sidechain_list = [aa_list[0]]
            aa_nterm_ionizable_nterm_list = [aa_list[0]]
            aa_cterm_ionizable_sidechain_list = [aa_list[-1]]
            aa_cterm_ionizable_cterm_list = []

        # capped N-terminus, sidechain still ionizable
        elif ionizable_cterm and not ionizable_nterm:
            aa_middle_list = aa_list[1:-1]
            aa_nterm_ionizable_sidechain_list = [aa_list[0]]
            aa_nterm_ionizable_nterm_list = []
            aa_cterm_ionizable_sidechain_list = [aa_list[-1]]
            aa_cterm_ionizable_cterm_list = [aa_list[-1]]

        # capped both N- and C-termimi, sidechain still ionizable
        elif not ionizable_cterm and not ionizable_nterm:
            aa_middle_list = aa_list[1:-1]
            aa_nterm_ionizable_sidechain_list = [aa_list[0]]
            aa_nterm_ionizable_nterm_list = []
            aa_cterm_ionizable_sidechain_list = [aa_list[-1]]
            aa_cterm_ionizable_cterm_list = []

        # Middle
        for aa in aa_middle_list:
            aa_position_in_sequence = "middle"
            (
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
            ) = self._get_aa_pkas_from_single_letter_aa(
                aa,
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
                pka_sets,
                aa_position_in_sequence,
            )

        # N-term, sidechains
        for aa in aa_nterm_ionizable_sidechain_list:
            aa_position_in_sequence = "nterm sidechains"
            (
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
            ) = self._get_aa_pkas_from_single_letter_aa(
                aa,
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
                pka_sets,
                aa_position_in_sequence,
            )

        # N-term, termini
        for aa in aa_nterm_ionizable_nterm_list:
            aa_position_in_sequence = "nterm nterm"
            (
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
            ) = self._get_aa_pkas_from_single_letter_aa(
                aa,
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
                pka_sets,
                aa_position_in_sequence,
            )

        # C-term, sidechains
        for aa in aa_cterm_ionizable_sidechain_list:
            aa_position_in_sequence = "cterm sidechains"
            (
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
            ) = self._get_aa_pkas_from_single_letter_aa(
                aa,
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
                pka_sets,
                aa_position_in_sequence,
            )

        # C-term, termini
        for aa in aa_cterm_ionizable_cterm_list:
            aa_position_in_sequence = "cterm cterm"
            (
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
            ) = self._get_aa_pkas_from_single_letter_aa(
                aa,
                base_pka_dict,
                acid_pka_dict,
                diacid_pka_dict,
                pka_sets,
                aa_position_in_sequence,
            )

        return base_pka_dict, acid_pka_dict, diacid_pka_dict



import numpy as np
import matplotlib.pyplot as plt

from itertools import cycle
from pichemist.config import PKA_SETS_NAMES
from pichemist.config import PLOT_LINE_WIDTHS as pld
from pichemist.config import REFERENCE_PKA_SET


def output_ph_q_curve(ph_q_dict, fig_filename):
    """Generates a pH/Q curve plot and saves it."""
    # Static parameters
    plt.matplotlib.rcParams.update({"font.size": 16})
    plt.figure(figsize=(8, 6))
    lines = ["-", "--", "-.", ":"]
    linew = [pld["w1"], pld["w1"],
             pld["w2"], pld["w2"],
             pld["w3"], pld["w2"],
             pld["w4"], pld["w4"]]
    linecycler = cycle(lines)
    linewcycler = cycle(linew)
    plt.ylabel("peptide charge")
    plt.xlabel("pH")
    plt.minorticks_on()
    plt.grid(True)

    # Plot curve for each pKa set
    i = 0
    for pka_set in PKA_SETS_NAMES:
        i += 1
        ph_q = ph_q_dict[pka_set]
        pl = plt.plot(ph_q[:, 0], ph_q[:, 1], next(linecycler),
                      label=pka_set, linewidth=next(linewcycler))
        if pka_set == REFERENCE_PKA_SET:
            plt.setp(pl, linewidth=8, linestyle="-", color="k")
        if i == 1:
            pH = ph_q[:, 0]
            q_m = ph_q[:, 1]
        else:
            q_m = np.column_stack([q_m, ph_q[:, 1]])
    plt.plot([7, 7], [np.min(q_m), np.max(q_m)], "k-")
    plt.plot(pH, pH*0, "k-")

    # Dynamic parameters
    plt.xlim([2, 12])
    plt.ylim([np.min(q_m), np.max(q_m)])
    plt.legend(loc="center right", bbox_to_anchor=[1.1, 0.5],
               ncol=1, shadow=True,
               fontsize=10).get_frame().set_alpha(1)
    plt.savefig(fig_filename)



import os
import subprocess
import tempfile

from pichemist.config import ACD_METHOD
from pichemist.config import PKA_LIMITS


class ACDPKaException(Exception):
    pass


class ACDPKaCalculator(object):
    """Uses ACD perceptabat to calculate pKa values."""

    def __init__(self):
        self.input_filepath = self._get_temp_filepath(".smi")
        self.output_filepath = self._get_temp_filepath(".out")
        self.pka_flag = self._get_pka_flag()

    def _get_pka_flag(self):
        return ACD_METHOD.value

    def _get_temp_filepath(self, suffix):
        """Gets a temporary filepath and closes the file."""
        file = tempfile.NamedTemporaryFile(suffix=suffix)
        path = file.name
        file.close()
        return path

    def _prepare_temp_input_file(self, smi_list):
        """Prepares the input file with SMILES."""
        with open(self.input_filepath, "w") as f:
            i = 0
            for smi in smi_list:
                i += 1
                f.write(f"{smi} tmpname{i}\n")

    def _get_temp_output_filepath(self):
        """Gets a temporary filepath and closes the file."""
        file = tempfile.NamedTemporaryFile(suffix=".out")
        path = file.name
        file.close()
        return path

    def _build_command(self):
        """Builds the perceptabat command to run pKa prediction."""
        return [
            "perceptabat",
            f"-TFNAME{self.output_filepath}",
            self.pka_flag,
            "-TPKA",
            self.input_filepath,
        ]

    def _get_status_output(self, *args, **kwargs):
        p = subprocess.Popen(*args, **kwargs)
        stdout, stderr = p.communicate()
        return p.returncode, stdout, stderr

    def _run_acd_exe(self, cmd):
        status, stdout, stderr = self._get_status_output(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        if status != 0:
            raise ACDPKaException(
                f"""Error while running the ACD subprocess:
                                  status: {status}
                                  command: {cmd}
                                  stdout: {stdout}
                                  stderr: {stderr}"""
            )

    def _check_existence_output(self):
        if not os.path.isfile(self.output_filepath):
            raise ACDPKaException("ACD did not generate the output file.")

    def _delete_temp_files(self):
        """Deletes the temp files of the instance."""
        for path in [self.input_filepath, self.output_filepath]:
            if os.path.exists(path):
                os.remove(path)

    def _parse_output(self, smi_list):
        """Parses the output of the software."""
        with open(self.output_filepath, "r") as f:
            base_pkas = list()
            acid_pkas = list()
            f.readline()  # skip first line
            for line in f.readlines():
                ln = line.split()
                mol_idx = int(ln[0])
                if "ACD_pKa_Apparent" in ln[1]:
                    pka = float(ln[2])
                if "ACD_pKa_DissType_Apparent" in ln[1]:
                    if ln[2] in ["MB", "B"]:
                        if pka > PKA_LIMITS["base_1"] and pka < PKA_LIMITS["base_2"]:
                            base_pkas.append((pka, smi_list[mol_idx - 1]))
                    if ln[2] in ["MA", "A"]:
                        if pka > PKA_LIMITS["acid_1"] and pka < PKA_LIMITS["acid_2"]:
                            acid_pkas.append((pka, smi_list[mol_idx - 1]))
        return (base_pkas, acid_pkas)

    def calculate_pka_from_list(self, smi_list):
        """Calculates the pKa values of a list of SMILES."""
        # Run the calculations
        self._prepare_temp_input_file(smi_list)
        cmd = self._build_command()
        self._run_acd_exe(cmd)
        self._check_existence_output()
        results = self._parse_output(smi_list)
        self._delete_temp_files()
        return results



from pichemist.fasta.matcher import FastaPKaMatcher


def _merge_pka_list(list_of_lists):
    """Merges a list of lists of pKa values."""
    PKA_VALUE_IDX = 0
    merged_pkas = list()
    for e in list_of_lists:
        for tup in e:
            merged_pkas.append(tup[PKA_VALUE_IDX])
    return merged_pkas


def _unpack_pka_list(pka_list):
    """Unpacks a list of pKa lists."""
    return pka_list[0], pka_list[1]


def merge_pkas_lists(base_pkas_list, acid_pkas_list):
    # Initialise the merged pKa lists
    base_pkas = dict()
    acid_pkas = dict()

    # Unpack the pKa values
    base_pkas_fasta, base_pkas_calc = _unpack_pka_list(base_pkas_list)
    acid_pkas_fasta, acid_pkas_calc = _unpack_pka_list(acid_pkas_list)

    # Merge FASTA and calculated values
    pka_sets_names = FastaPKaMatcher().get_pka_sets_names()
    for pka_set in pka_sets_names:
        base_pkas[pka_set] = _merge_pka_list([base_pkas_fasta[pka_set], base_pkas_calc])
        acid_pkas[pka_set] = _merge_pka_list([acid_pkas_fasta[pka_set], acid_pkas_calc])
    return base_pkas, acid_pkas



from setuptools import find_packages
from setuptools import setup

__doc__ = (
    "Calculates the isoelectric point (pI) of a peptide/protein "
    "using the Henderson-Hasselbalch equation with the predicted "
    "pKa values of its amino acids."
    ""
)

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="pichemist",
    packages=find_packages(),
    version="0.2.0",
    description=__doc__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Gian Marco Ghiandoni",
    author_email="ghiandoni.g@gmail.com",
    license="Apache 2.0",
    install_requires=["matplotlib", "rdkit", "biopython"],
    setup_requires=["pytest-runner", "flake8"],
    tests_require=["pytest==7.1.2"],
    test_suite="test",
    package_dir={"pichemist": "pichemist"},
    include_package_data=True,
    url="https://github.com/AstraZeneca/peptide-tools/tree/master/pIChemiSt",
    entry_points={
        "console_scripts": ["pichemist=pichemist.cli:main"],
    },
)



from rdkit import Chem


def _pattern_match_rdkit(smiles, smarts):
    """RDKit SMARTS matching counter."""
    mol = Chem.MolFromSmiles(smiles)
    mol = Chem.AddHs(mol)
    pat = Chem.MolFromSmarts(smarts)
    n = 0
    for _ in mol.GetSubstructMatches(pat):
        n += 1
    return n


def pattern_match(smiles, smarts):
    """Interface for SMARTS matching."""
    return _pattern_match_rdkit(smiles, smarts)



import json

from helpers import examples_dir
from pichemist.fasta.smarts import AA_SMARTS_SET
from pichemist.fasta.pka_sets import FASTA_PKA_SETS
from pichemist.smarts.pka_set import SS_SMARTS_PKA_SET


def test_fasta_pka_sets():
    with open(f"{examples_dir}/fasta_pka_sets_expected.json") as f:
        expected = json.load(f)
    assert expected == FASTA_PKA_SETS


def test_ss_smarts_pka_set():
    with open(f"{examples_dir}/ss_smarts_pka_set_expected.json") as f:
        expected = json.load(f)
    assert expected == SS_SMARTS_PKA_SET


def test_aa_smarts_pka_set():
    with open(f"{examples_dir}/aa_smarts_pka_set_expected.json") as f:
        expected = json.load(f)
    assert expected == AA_SMARTS_SET



import base64
from enum import Enum
from io import BytesIO

from pichemist.utils import get_logger
from rdkit import Chem
from rdkit.Chem import Draw
from rdkit.Chem import rdDepictor
from rdkit.Chem.MolStandardize import rdMolStandardize

log = get_logger(__name__)


class Atoms(Enum):
    CARBON = Chem.Atom(6)
    OXYGEN = Chem.Atom(8)


class MolStandardiser(object):
    """Deals with molecule standardisation."""

    def __init__(self):
        pass

    def neutralise_molecule(self, mol):
        """
        Replace all ionized centers by the corresponsing
        neutral form. Used to track constanty ionized
        fragments, like quaternary amines.
        https://www.rdkit.org/docs/Cookbook.html

        """
        ION_SMARTS = (
            "[+1!h0!$([*]~[-1,-2,-3,-4]),-1!$([*]~[+1,+2,+3,+4])]"  # noqa: E501
        )

        pattern = Chem.MolFromSmarts(ION_SMARTS)
        at_matches = mol.GetSubstructMatches(pattern)
        at_matches_list = [y[0] for y in at_matches]
        if len(at_matches_list) > 0:
            for at_idx in at_matches_list:
                atom = mol.GetAtomWithIdx(at_idx)
                chg = atom.GetFormalCharge()
                hcount = atom.GetTotalNumHs()
                atom.SetFormalCharge(0)
                atom.SetNumExplicitHs(hcount - chg)
                atom.UpdatePropertyCache()
        return mol

    def standardise_molecule(self, mol):
        """
        Runs molecule standardisation using a built-in
        method and some SMARTS-based heuristics.

        """
        clean_mol = rdMolStandardize.Cleanup(mol)
        m = self.neutralise_molecule(clean_mol)
        return m


class PeptideCutter(object):
    """Deals with splitting and capping of peptides."""

    def __init__(self):
        pass

    @staticmethod
    def _add_acetyl_to_mol(rw_mol, atom_attach):
        """Adds a C(=O)(C) to an atom in a RW molecule."""
        c_carbonyl_idx = rw_mol.AddAtom(Atoms.CARBON.value)
        o_carbonyl_idx = rw_mol.AddAtom(Atoms.OXYGEN.value)
        methyl_idx = rw_mol.AddAtom(Atoms.CARBON.value)
        rw_mol.AddBond(atom_attach, c_carbonyl_idx, Chem.BondType.SINGLE)
        rw_mol.AddBond(c_carbonyl_idx, o_carbonyl_idx, Chem.BondType.DOUBLE)
        rw_mol.AddBond(c_carbonyl_idx, methyl_idx, Chem.BondType.SINGLE)
        return rw_mol

    def _fragment_and_cap(self, mol, amide_atoms):
        """
        Function for breaking and capping.
        Creates a RW mol; iterates through a set of breakable
        atom indices; for each pair of indices, splits the
        molecule in two parts and adds an acetyl capping
        to the first atom (nitrogen of the amine NH), and a carbon
        to the second atom (carbon of the carbonyl C=O). This
        ensures that both fragments are capped with acetyles.

        """
        rw_mol = Chem.RWMol(mol)
        for atoms in amide_atoms:
            amine_idx = atoms[0]
            carbonyl_idx = atoms[1]
            # Break bond
            rw_mol.RemoveBond(amine_idx, carbonyl_idx)
            # Add acetyl to the amine
            rw_mol = PeptideCutter._add_acetyl_to_mol(rw_mol, amine_idx)
            # Add carbon to the carbonyl
            c_idx = rw_mol.AddAtom(Atoms.CARBON.value)
            rw_mol.AddBond(carbonyl_idx, c_idx, Chem.BondType.SINGLE)
        return Chem.MolToSmiles(rw_mol)

    def break_amide_bonds_and_cap(self, mol):
        """
        Breaks the amine bonds in a molecule and
        caps them with acetyl groups.

        """
        # Secondary and tertiary amide bonds
        AMIDE_SMARTS = "[NX3,NX4;H0,H1][CX3](=[OX1])"
        amide_pattern = Chem.MolFromSmarts(AMIDE_SMARTS)

        amide_atoms = list()
        smiles = Chem.MolToSmiles(mol)
        if mol.HasSubstructMatch(amide_pattern):
            amide_matches = mol.GetSubstructMatches(amide_pattern)
            for atoms in amide_matches:
                amine_idx = atoms[0]
                carbonyl_idx = atoms[1]
                amide_atoms.append((amine_idx, carbonyl_idx))
            smiles = self._fragment_and_cap(mol, amide_atoms)

        smiles_list = smiles.split(".")
        log.debug(f"Obtained {len(smiles_list)} fragments")
        return smiles_list


def smiles_to_image(smiles, b64encode=True):
    mol = Chem.MolFromSmiles(smiles)
    rdDepictor.Compute2DCoords(mol)
    img = Draw.MolToImage(mol, kekulize=True)
    buffered = BytesIO()
    img.save(buffered, format="PNG")
    img_byte_data = buffered.getvalue()
    if b64encode:
        base64_image = base64.b64encode(img_byte_data).decode("utf-8")
    return base64_image
